from tktl.core import ExtendedEnum


class EndpointKinds(str, ExtendedEnum):
    BASIC = "basic"
    TYPED = "typed"
    ARROW = "arrow"
    PROFILED = "profiled"


class ProfileKinds(str, ExtendedEnum):
    BINARY = "binary"
    REGRESSION = "regression"



import typing as t

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from pydantic import BaseModel

from .endpoints import ArrowEndpoint, BasicEndpoint, ProfiledEndpoint, TypedEndpoint

# Enpoint types sorted by precedence
ENDPOINT_TYPES = [ProfiledEndpoint, ArrowEndpoint, TypedEndpoint, BasicEndpoint]


class Tktl:
    def __init__(self):
        self._endpoints = []

    @property
    def endpoints(self):
        return self._endpoints

    def endpoint(
        self,
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        profile: t.Optional[str] = None,
        profile_columns: t.Optional[t.List[str]] = None,
        **kwargs,
    ):

        try:
            Constructor = next(
                endpoint
                for endpoint in ENDPOINT_TYPES
                if endpoint.supported(X=X, y=y, profile=profile)
            )
        except StopIteration:
            raise ValueError(
                f"Arguments of type X={type(X)} y={type(y)} and profile={profile}"
                " are not supported by any endpoint type."
            )

        def decorator(f):
            endpoint = Constructor(name=f.__name__, func=f, **kwargs, X=X, y=y)
            self._endpoints.append(endpoint)
            return f

        return decorator



import os
import pkgutil
import sys
from typing import Optional

from tktl import __version__
from tktl.core.loggers import LOG
from tktl.core.managers.base import BaseConfigManager
from tktl.core.managers.constants import TEMPLATE_PROJECT_DIRS, TEMPLATE_PROJECT_FILES
from tktl.core.validation.project import validate_project_contents


class ProjectManager(BaseConfigManager):
    """Manages project configuration tktl.yaml file."""

    TKTL_DIR: Optional[str] = None
    CONFIG_FILE_NAME: str = "tktl.yaml"
    CONFIG: Optional[dict] = None

    @classmethod
    def set_path(cls, path: str) -> None:
        cls.TKTL_DIR = path

    @classmethod
    def init(cls, path: str, name: str) -> None:
        cls.create_dir(path)
        cls.set_path(path)
        cls.init_config(name)
        cls.create_template()

    @classmethod
    def init_project(cls, path: Optional[str], name: str) -> str:
        if not path:
            path = os.path.expanduser(".")
        project_path = os.path.join(path, name)
        if os.path.isdir(project_path):
            cls.safe_init(project_path, name)
        else:
            cls.init(project_path, name)
        return project_path

    @classmethod
    def safe_init(cls, path: str, name: str) -> None:
        LOG.log(
            "WARNING: Path specified already exists. "
            "Any configuration will be overwritten. Proceed?[Y/n]",
            color="yellow",
        )
        overwrite = input()
        if overwrite == "Y":
            cls.init(path, name)
        elif overwrite == "n":
            sys.exit()
        else:
            cls.safe_init(path, name)

    @classmethod
    def create_template(cls) -> None:
        cls.create_subdirs()
        cls.copy_dist_files()

    @classmethod
    def copy_dist_files(cls) -> None:
        for file in TEMPLATE_PROJECT_FILES:
            data = pkgutil.get_data("tktl", file)
            cls._copy_file(file, data)

    @classmethod
    def create_subdirs(cls) -> None:
        for path in TEMPLATE_PROJECT_DIRS:
            full_path = os.path.join(cls.TKTL_DIR, path) if cls.TKTL_DIR else path
            os.makedirs(full_path, exist_ok=True)

    @classmethod
    def _copy_file(cls, file, data, replace_path=""):
        file_path = file.replace("template/", replace_path)
        try:
            with open(os.path.join(cls.TKTL_DIR, file_path), "w") as f:
                f.write(data.decode("utf-8").replace("__version__", __version__))
        except UnicodeDecodeError:
            with open(os.path.join(cls.TKTL_DIR, file_path), "wb") as f:
                f.write(data)

    @classmethod
    def validate_project_config(cls, path="."):
        return validate_project_contents(path=path)



from typing import Optional

from pydantic import UUID4, BaseModel


class TaktileUser(BaseModel):
    id: UUID4
    email: Optional[str] = None
    is_active: bool
    is_superuser: bool
    username: str
    full_name: Optional[str] = None
    source_id: Optional[UUID4] = None



from tktl import __version__
from tktl.commands import BaseTaktileApiCommand, CommandBase
from tktl.core.clients.http_client import interpret_response
from tktl.core.config import settings
from tktl.core.exceptions.exceptions import APIClientException
from tktl.core.loggers import LOG
from tktl.core.schemas.user import TaktileUser
from tktl.login import login, logout, set_api_key
from tktl.version import TaktileVersionChecker


class LogInCommand(BaseTaktileApiCommand):
    def execute(self, api_key=None):
        if login(api_key):
            response = self.client.get(f"{settings.API_V1_STR}/users/me")
            try:
                user = interpret_response(response, TaktileUser)
            except APIClientException as e:
                LOG.error(f"Authentication failed: {e.detail}")
                return False
            LOG.log(
                f"Authentication successful for user: {user.username}", color="green"
            )
            return True
        return False


class LogOutCommand(CommandBase):
    def execute(self):
        logout()


class ShowVersionCommand(CommandBase):
    def execute(self, check: bool = False):
        LOG.log(__version__)
        if check:
            TaktileVersionChecker.look_for_new_version()


class SetApiKeyCommand(CommandBase):
    def execute(self, api_key):
        if not api_key:
            LOG.error("API Key cannot be empty.")
            return
        set_api_key(api_key=api_key)



import os
import shutil

import pytest

from tktl.commands.init import init_project
from tktl.commands.validate import validate_project_config
from tktl.core.managers.constants import TEMPLATE_PROJECT_DIRS, TEMPLATE_PROJECT_FILES
from tktl.core.managers.project import ProjectManager


def test_create_in_existing_directory(monkeypatch, tmpdir):
    monkeypatch.setattr("builtins.input", lambda: "Y")
    ProjectManager.init_project(tmpdir.dirpath(), "test_project_dir")

    for d in TEMPLATE_PROJECT_DIRS:
        assert os.path.isdir(os.path.join(tmpdir.dirpath(), "test_project_dir", d))
    for file in TEMPLATE_PROJECT_FILES:
        basename = file.replace("template/", "")
        assert os.path.exists(
            os.path.join(tmpdir.dirpath(), "test_project_dir", basename)
        )
    assert os.path.exists(
        os.path.join(tmpdir.dirpath(), "test_project_dir", "tktl.yaml")
    )


def test_create_project(create_proj):
    for file in TEMPLATE_PROJECT_FILES:
        basename = file.replace("template/", "")
        assert os.path.exists(os.path.join("sample_project", basename))
    assert os.path.exists(os.path.join("sample_project", "tktl.yaml"))

    for d in TEMPLATE_PROJECT_DIRS:
        assert os.path.isdir(os.path.join("sample_project", d))


def test_safe_init(monkeypatch, caplog):
    monkeypatch.setattr("builtins.input", lambda: "n")
    ProjectManager.init_project(None, "sample_project")

    with pytest.raises(SystemExit):
        ProjectManager.safe_init(None, "sample_project")
        assert "Path specified already exists" in caplog.text
    shutil.rmtree("sample_project")


def test_overwrite(monkeypatch, capsys):
    monkeypatch.setattr("builtins.input", lambda: "Y")
    ProjectManager.init_project(None, "sample_project")
    init_project(None, "sample_project")
    out, err = capsys.readouterr()
    assert "Project scaffolding created successfully" in out
    shutil.rmtree("sample_project")


def test_validate(capsys):
    ProjectManager.init_project(None, "sample_project")
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert out == "Project scaffolding is valid!\n"
    shutil.rmtree("sample_project")

    ProjectManager.init_project(None, "sample_project")
    shutil.rmtree("sample_project/src")
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert err is not None
    assert "Project scaffolding is invalid: Missing Files or Directory ❌\n" in out
    assert "Missing directories in repository: src" in err
    shutil.rmtree("sample_project")

    ProjectManager.init_project(None, "sample_project")
    with open("sample_project/tktl.yaml", "w") as inf:
        inf.write(
            """
default_branch_name: master
service:
  replicas: 3
  compute_type: CPUqqqqqqqq
  endpoint_type: RESTuuuuuu
"""
        )
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert "Project scaffolding is invalid: Invalid Config File ❌\n" == out
    shutil.rmtree("sample_project")



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Sequence, Tuple, Type, _GenericAlias  # type: ignore

import numpy  # type: ignore
import pandas  # type: ignore
from cached_property import cached_property  # type: ignore
from pydantic import BaseModel
from pydantic.main import ModelMetaclass

from tktl.core.loggers import LOG
from tktl.core.serializers import to_pydantic
from tktl.core.serializers.base import CustomDeserializingModelT
from tktl.core.t import RestSchemaTypes
from tktl.core.utils import DelayedLoader, check_and_get_value
from tktl.registration.validation import (
    data_frame_convertible,
    input_pandas_representation,
    output_pandas_representation,
    series_convertible,
)

_SUPPORTS_ARROW_SAMPLE_DATA = [
    DelayedLoader,
    pandas.DataFrame,
    pandas.Series,
    numpy.ndarray,
    Sequence,
    dict,
]
_SUPPORTS_REST_SAMPLE_DATA = _SUPPORTS_ARROW_SAMPLE_DATA + [ModelMetaclass]
_SUPPORTED_TYPES = _SUPPORTS_REST_SAMPLE_DATA + [_GenericAlias]


def check_supported_inputs(value, types):
    return any([isinstance(value, t_) for t_ in types])


class EndpointSchema(ABC):
    KIND: str

    def __init__(
        self,
        value: Any,
        endpoint_kind: str,
        endpoint_name: str,
        user_defined_model: Type[BaseModel] = None,
    ):
        self.endpoint_kind = endpoint_kind
        self.value = value
        self.is_supported = check_supported_inputs(self.value, _SUPPORTED_TYPES)
        self.supports_rest_sample_data = check_supported_inputs(
            self.value, _SUPPORTS_REST_SAMPLE_DATA
        )
        self.supports_arrow_sample_data = check_supported_inputs(
            value, _SUPPORTS_ARROW_SAMPLE_DATA
        )
        if not self.is_supported:
            LOG.warning(
                f"Endpoint '{endpoint_name}': schema discovery not supported for inputs of type {type(value)}"
            )
        if user_defined_model:
            value = user_defined_model

        value = check_and_get_value(value, first_batch=True)
        self.pydantic: Type[CustomDeserializingModelT] = to_pydantic(
            value, unique_id=f"{endpoint_name}__{self.KIND}"
        )

    @cached_property
    @abstractmethod
    def pandas_convertible(self):
        raise NotImplementedError


class EndpointOutputSchema(EndpointSchema):
    KIND = "output"

    @cached_property
    def pandas_convertible(self):
        return series_convertible(self.value)

    @cached_property
    def names(self):
        if self.pandas_convertible:
            return self.pandas_representation.name
        else:
            return get_model_names(self.pydantic)

    @cached_property
    def pandas_representation(self):
        if self.pandas_convertible:
            return output_pandas_representation(self.value)
        else:
            return None


class EndpointInputSchema(EndpointSchema):
    KIND = "input"

    @cached_property
    def pandas_convertible(self):
        value = check_and_get_value(self.value, first_batch=True)
        return data_frame_convertible(value)

    @cached_property
    def names(self):
        if self.pandas_convertible:
            return list(self.pandas_representation.columns)
        else:
            return get_model_names(self.pydantic)

    @cached_property
    def pandas_representation(self):
        if self.pandas_convertible:
            return input_pandas_representation(self.value)
        else:
            return None


def get_model_names(model: CustomDeserializingModelT):
    schema = model.schema()
    if schema["title"] == RestSchemaTypes.DICT.value:
        return list(schema["properties"].keys())
    elif (
        schema["title"] == RestSchemaTypes.SEQUENCE.value
        or schema["title"] == RestSchemaTypes.ARRAY.value
    ):
        return get_nested_type_names(schema)
    else:
        raise ValueError(f"Schema named: {schema['title']} is not recognized")


def get_nested_type_names(schema) -> List[str]:
    schema, inner_type = walk_sequence_schema(schema)
    n_features = schema[0]
    if len(schema) == 1:
        shapes = ""
    else:
        shapes_str = "x".join([str(dim) for dim in schema[1:]])
        shapes = f" (Shape: {shapes_str})"
    return [f"Feature {i}{shapes}" for i in range(n_features)]


def walk_sequence_schema(schema: Dict, shapes=()) -> Tuple[Tuple[int], str]:
    if schema["type"] == "array":
        shapes += (schema["maxItems"],)
        return walk_sequence_schema(schema["items"], shapes=shapes)
    else:
        return shapes, schema["type"]



import abc
import os
from queue import Queue
from threading import Thread
from typing import Generator, List, Tuple

from click import secho
from docker.models.containers import Container  # type: ignore

from tktl.core.config import settings


def set_verbosity(verbose):
    set_logger_verbosity(verbose)
    set_grpc_verbosity(verbose)


def set_grpc_verbosity(verbose):
    if verbose > 1:
        os.environ["GRPC_TRACE"] = "channel,http"
        os.environ["GRPC_VERBOSITY"] = "debug"
    else:
        os.environ.pop("GRPC_TRACE", None)
        os.environ.pop("GRPC_VERBOSITY", None)


def set_logger_verbosity(verbose):
    if verbose == 0:
        if settings.VERBOSE:
            verbose += 1
        elif settings.DEBUG:
            verbose += 2
    if verbose >= 1 or settings.VERBOSE:
        LOG.VERBOSE = True
    if verbose > 1 or settings.DEBUG:
        LOG.DEBUG = True


class Logger(abc.ABC):
    VERBOSE = False
    DEBUG = False

    @abc.abstractmethod
    def log(self, msg, *args, **kwargs):
        pass

    @abc.abstractmethod
    def warning(self, msg, *args, **kwargs):
        pass

    @abc.abstractmethod
    def error(self, msg, *args, **kwargs):
        pass

    def debug(self, msg, *args, **kwargs):
        pass

    def trace(self, msg, *args, **kwargs):
        pass


class MuteLogger(Logger):
    def log(self, msg, *args, **kwargs):
        pass

    def warning(self, msg, *args, **kwargs):
        pass

    def error(self, msg, *args, **kwargs):
        pass

    def trace(self, msg, *args, **kwargs):
        pass


class CliLogger(Logger):
    @staticmethod
    def _log(message, color=None, err=False):
        message = str(message)
        color = color if settings.USE_CONSOLE_COLORS else None
        secho(message, fg=color, err=err)

    def trace(self, message, color=None, err=False):
        if self.VERBOSE:
            self._log(message, color=color, err=err)

    def log(self, message, color=None, err=False):
        self._log(message, color=color, err=err)

    def error(self, message, *args, **kwargs):
        color = "red" if settings.USE_CONSOLE_COLORS else None
        self._log(message, color=color, err=True)

    def warning(self, message, *args, **kwargs):
        color = "yellow" if settings.USE_CONSOLE_COLORS else None
        self._log(message, color=color)

    def debug(self, message, *args, **kwargs):
        if self.DEBUG:
            self._log("DEBUG: {}".format(message))


def stream_blocking_logs(
    arrow_container: Container, rest_container: Container, color_logs: bool
) -> None:
    arrow_stream = blocking_generator(
        arrow_container.logs(stream=True), kind="arrow", color_logs=color_logs
    )
    rest_stream = blocking_generator(
        rest_container.logs(stream=True), kind="rest", color_logs=color_logs
    )
    streams = [arrow_stream, rest_stream]

    try:
        for msg, color in merge_blocking_iterators(streams):
            LOG.log(message=msg, color=color)
    except (KeyboardInterrupt, StopIteration, RuntimeError):
        LOG.log("Exiting...", color="red")
        rest_container.kill()
        arrow_container.kill()


def blocking_generator(
    container_logs: Generator[bytes, None, None],
    kind: str = "arrow",
    color_logs: bool = True,
) -> Generator[Tuple, None, None]:
    while True:
        try:
            line = container_logs.__next__()
        except StopIteration:
            return
        if kind == "arrow":
            yield f"> {line.decode()}".strip(), "bright_magenta" if color_logs else None
        else:
            yield f"> {line.decode()}".strip(), "bright_green" if color_logs else None
        if not line:
            return


def merge_blocking_iterators(
    generators: List[Generator[Tuple, None, None]]
) -> Generator[Tuple[str, str], None, None]:

    queue: Queue = Queue()

    def consume(_queue: Queue, stream: Generator) -> None:
        for i in stream:
            _queue.put(i)

    for generator in generators:
        thread = Thread(target=consume, args=(queue, generator))
        thread.start()

    while True:
        out = queue.get()
        if not out:
            return
        else:
            yield out


LOG = CliLogger()
MUTE_LOG = MuteLogger()



import os

import pytest

from tktl.core.clients import DeploymentApiClient
from tktl.core.config import settings
from tktl.core.exceptions import TaktileSdkError
from tktl.core.managers.auth import AuthConfigManager


def test_instantiate_taktile_client():
    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)
    client = DeploymentApiClient()
    with pytest.raises(TaktileSdkError):
        client.get_endpoint_by_name("admin/df.asfm", "", "")
    assert "X-Api-Key" in client.default_headers
    assert client.default_headers["X-Api-Key"] == key
    assert (
        client.get_path(f"{settings.API_V1_STR}/deployments")
        == f"{settings.DEPLOYMENT_API_URL}{settings.API_V1_STR}/deployments"
    )



from pandas.api.types import (
    is_categorical_dtype,
    is_datetime64_any_dtype,
    is_float_dtype,
    is_integer_dtype,
)
from src.endpoints import X


def test_schema(json_metadata):
    typechecks = {
        "Loan Amount": is_integer_dtype,
        "Funded Amount": is_integer_dtype,
        "Term": is_integer_dtype,
        "Employment Length": is_categorical_dtype,
        "Home Ownership": is_categorical_dtype,
        "Annual Income": is_float_dtype,
        "Issue Date": is_datetime64_any_dtype,
        "Purpose": is_categorical_dtype,
        "Debt-to-Income Ratio": is_float_dtype,
    }

    json_metadata["section"] = "Data"
    json_metadata["pass_message"] = "Column names and column types are correct"
    json_metadata["fail_message"] = "Incorrect schema"

    # assert correct columns are present
    assert set(typechecks.keys()) == set(X.columns)

    # assert correct column types
    for col, typecheck in typechecks.items():
        assert typecheck(X[col])



TEMPLATE_PROJECT_DIRS = [
    "src/",
    "assets/",
    "tests/",
]

TEMPLATE_PROJECT_FILES = [
    "template/src/endpoints.py",
    "template/assets/model.joblib",
    "template/assets/loans_test.pqt",
    "template/tests/test_accuracy.py",
    "template/tests/test_plausibility.py",
    "template/tests/test_schema.py",
    "template/tests/test_taktile_endpoint_tests.py",
    "template/.gitignore",
    "template/.gitattributes",
    "template/.buildfile",
    "template/.dockerignore",
    "template/README.md",
    "template/requirements.txt",
]


CONFIG_FILE_TEMPLATE = """
# If this option is set only commits with a commit message starting
# by the deployment_prefix will be deployed
# deployment_prefix: "#deploy"
# undeployment_prefix: "#undeploy"

service:

  # For the `instance_type` format, use the following: <instance_kind>.<instance_size>
  # Only 'gp' (general purpose) instance_kind is supported for now
  # instance_size can be: small, medium, large, xlarge, or xxlarge.
  # This is the instance type that will be used for both arrow and rest

  # Rest deployment scaling parameters
  rest:
    instance_type: gp.small
    # Maximum and minimum number of replicas to which the REST service is allowed to reach
    # by the auto-scaler
    max_replicas: 1
    # Minimum, or starting number of replicas
    replicas: 1

  # Arrow deployment scaling parameters
  arrow:
    instance_type: gp.small
    # Arrow # replicas are fixed and don't autoscale
    replicas: 1

version: {version}
"""



from enum import Enum
from typing import List

import pandas
from numpy.testing import assert_array_equal
from pydantic import BaseModel

from tktl.core.serializers import to_pydantic
from tktl.core.serializers.rest import (
    DataFrameSerializer,
    PassThroughSerializer,
    SeriesSerializer,
)


def test_frame_serializer(serializer_df_inputs: pandas.DataFrame):
    model = to_pydantic(serializer_df_inputs)
    as_pydantic_model = DataFrameSerializer.serialize(
        serializer_df_inputs, output_model=model
    )
    assert isinstance(as_pydantic_model, list)
    assert as_pydantic_model[0].index == serializer_df_inputs.index.values[0]
    serialized = model.parse_obj(serializer_df_inputs.to_dict(orient="records"))
    deser = serialized.deserialize()
    assert isinstance(deser, pandas.DataFrame)
    assert list(deser.columns) == list(serializer_df_inputs.columns)
    assert_array_equal(deser["A"].values, serializer_df_inputs["A"].values)
    assert set(serializer_df_inputs.columns.tolist()).issubset(
        set(serialized.dict()["__root__"][0].keys())
    )


def test_series_serializer(serializer_series_inputs: pandas.Series):
    model = to_pydantic(serializer_series_inputs)
    as_pydantic_model = SeriesSerializer.serialize(
        serializer_series_inputs, output_model=model
    )
    assert isinstance(as_pydantic_model, BaseModel)
    assert isinstance(as_pydantic_model.dict()["B"], list)


def test_custom_serializer():
    class Transformation(str, Enum):
        upper = "upper"
        lower = "lower"

    class Payload(BaseModel):
        data: List[str]
        transformation: Transformation

    class Response(BaseModel):
        prediction: List[str]
        transformation: str

    model = to_pydantic(Payload)
    as_pydantic_model = PassThroughSerializer.serialize(
        value=model, output_model=Response
    )
    assert as_pydantic_model == model



import os
import time
from typing import Dict

import docker  # type: ignore

from tktl.commands.health import GetGrpcHealthCommand, GetRestHealthCommand
from tktl.core.exceptions.exceptions import APIClientException, MissingDocker
from tktl.core.loggers import LOG, MUTE_LOG

TESTING_DOCKERFILE = "Dockerfile.taktile-cli-testing"
MULTI_STAGE_BUILD_STEP_NAME = "build_step"


class DockerManager:
    def __init__(self, path):
        try:
            self._client = docker.from_env()
            self._path = path
        except docker.errors.DockerException as err:
            raise MissingDocker from err

    def get_docker_file(self) -> str:
        with open(os.path.join(self._path, ".buildfile")) as fp:
            return fp.read()

    def stream_logs(self, container) -> None:
        for line in container.logs(stream=True):
            LOG.log(f"> {line.decode()}".strip())

    def _patch_docker_file(
        self, input_docker_file: str = ".buildfile", output: str = TESTING_DOCKERFILE
    ):
        """patch_docker_file

        Remove kaniko related config
        """
        with open(os.path.join(self._path, input_docker_file)) as fp:
            lines = fp.readlines()
            desired_contents = []
            for line in lines:
                if line.startswith("ARG RESOURCE_NAME"):
                    break
                desired_contents.append(line)

        with open(os.path.join(self._path, output), "w") as fp:
            fp.writelines(desired_contents)

    @staticmethod
    def _is_multi_stage(file_path: str) -> bool:
        with open(file_path) as fp:
            lines = fp.readlines()
            return any(
                x.startswith(f"FROM {MULTI_STAGE_BUILD_STEP_NAME}") for x in lines
            )

    def _remove_patched_docker_file(self, file_path: str = TESTING_DOCKERFILE):
        os.remove(os.path.join(self._path, file_path))

    def remove_image(self, image):
        self._client.images.remove(image, force=True)

    def build_image(
        self, dockerfile: str, nocache: bool = False, buildargs: Dict = None
    ) -> str:

        if nocache:
            try:
                image = self._client.images.get("taktile-cli-test:latest")
                self._client.images.remove(image.id)
            except docker.errors.ImageNotFound:
                pass

        # TODO: After phasing out support for non multistage docker files, remove this.
        if DockerManager._is_multi_stage(dockerfile):
            LOG.log(
                f"Multistage dockerfile, building target {MULTI_STAGE_BUILD_STEP_NAME}..."
            )
            image, stream = self._client.images.build(
                path=self._path,
                dockerfile=dockerfile,
                tag="taktile-cli-test",
                nocache=nocache,
                target=MULTI_STAGE_BUILD_STEP_NAME,
                buildargs=buildargs,
                forcerm=True,
            )
        else:
            LOG.log("Patching dockerfile...")
            self._patch_docker_file(
                input_docker_file=dockerfile, output=TESTING_DOCKERFILE
            )
            image, stream = self._client.images.build(
                path=self._path,
                dockerfile=TESTING_DOCKERFILE,
                tag="taktile-cli-test",
                nocache=nocache,
                forcerm=True,
            )
            self._remove_patched_docker_file()

        for chunk in stream:
            if "stream" in chunk:
                for line in chunk["stream"].splitlines():
                    LOG.trace("> " + line)

        return image.id

    def test_import(self, image_id: str):
        try:
            container = self._client.containers.run(
                image_id, "python -c 'from src.endpoints import tktl'", detach=True,
            )
            self.stream_logs(container)

            status = container.wait()
            return status, container
        finally:
            container.remove()

    def test_unittest(self, image_id: str):
        try:
            container = self._client.containers.run(
                image_id, "python -m pytest ./user_tests/", detach=True
            )
            self.stream_logs(container)

            status = container.wait()
            return status, container
        finally:
            container.remove()

    def run_rest_container(
        self, image_id: str, detach: bool = True, auth_enabled: bool = True
    ):
        return self._client.containers.run(
            image_id,
            detach=detach,
            entrypoint="/start-rest.sh",
            environment={"AUTH_ENABLED": auth_enabled},
            ports={"80/tcp": 8080},
            remove=True,
            stderr=True,
            stdout=True,
        )

    def run_arrow_container(
        self, image_id: str, detach: bool = True, auth_enabled: bool = True
    ):
        return self._client.containers.run(
            image_id,
            detach=detach,
            entrypoint="/start-flight.sh",
            environment={"AUTH_ENABLED": auth_enabled},
            ports={"5005/tcp": 5005},
            remove=True,
            stderr=True,
            stdout=True,
        )

    def run_containers(
        self, image_id: str, detach: bool = True, auth_enabled: bool = True
    ):
        arrow_container = self.run_arrow_container(
            image_id=image_id, detach=detach, auth_enabled=auth_enabled
        )
        rest_container = self.run_rest_container(
            image_id=image_id, detach=detach, auth_enabled=auth_enabled
        )
        return arrow_container, rest_container

    def run_profiling_container(self):
        try:
            container = self._client.containers.run(
                "taktile/taktile-profiler:latest",
                entrypoint="profiler profile -l",
                network_mode="host",
                detach=True,
            )
            self.stream_logs(container)
            status = container.wait()
            return status, container
        finally:
            container.remove()

    def run_and_check_health(
        self,
        image_id: str,
        kill_on_success: bool = False,
        auth_enabled: bool = True,
        timeout: int = 7,
        retries: int = 7,
    ):
        arrow_container, rest_container = self.run_containers(
            image_id=image_id, detach=True, auth_enabled=auth_enabled
        )
        grpc_health_cmd = GetGrpcHealthCommand(
            branch_name="",
            repository="",
            local=True,
            logger=MUTE_LOG if not LOG.VERBOSE else LOG,
            skip_auth=True,
        )
        rest_health_cmd = GetRestHealthCommand(
            branch_name="",
            repository="",
            local=True,
            logger=MUTE_LOG if not LOG.VERBOSE else LOG,
            skip_auth=True,
        )

        try:
            for _ in range(retries):
                try:
                    time.sleep(timeout)
                    rest_response = rest_health_cmd.execute()
                    grpc_response = grpc_health_cmd.execute()
                    return rest_response, grpc_response, arrow_container, rest_container
                except (APIClientException, Exception):
                    pass

            return None, None, arrow_container, rest_container
        finally:
            if kill_on_success:
                arrow_container.kill()
                rest_container.kill()
